#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GRLayout.py
#
# DESCRIPTION:
# Classes related to the Layout section
#
# NOTES:
#


from gnue.common.GObjects import *
from gnue.common import GDebug, GConditions
from gnue.common.GTrigger import GTrigger,GTriggerExtension

from GRParameters import GRStubParam
import GRExceptions
import string
from xml.sax import saxutils


############################################################
#
# Used to send structural comment
# Chose not to do  "if COMMENT == 1: dest.write"
# for performance reasons  # TODO: Was this a good decision?
#
def _structuralComment(dest, text):
  dest.write(text)

def _noStructuralComment(dest, text):
  pass

structuralComment = _noStructuralComment


############################################################
#
# GRLayout
#
class GRLayout (GObj):
  def __init__(self, parent):
    GObj.__init__(self, parent, type='GRLayout')
    self._inits = (self.initialize,)


  def initialize(self):
    # Find the xml namespace in use by any child passthru objects
    # If there is more than one object attached to a layout
    # object, then create an unbound section to contain them.
    # This is a convenience for the layout engine code :)

    if  len(self._children) > 1 or \
       ( len(self._children) and \
         self._children[0]._type != 'GRSection' ):

      temp = self._children[:]
      self._children = []
      section = GRSection(self)
      section._children = temp
      section._buildObject()
    elif len(self._children) == 0:
      raise GRExceptions.NoLayoutDefined, \
        "The layout section does not contain any instructions. What do I do?"

    self._xmlchildnamespace = self._findNamespace(self)
    self._parent._namespace = self._xmlchildnamespace


  # Find the xml namespace in use by any child passthru objects
  def _findNamespace(self, object):
    for child in object._children:
      try:
        if child._xmlnamespace:
          namespace =  child._xmlnamespace
        else:
          namespace = self._findNamespace(child)
        if namespace:
          return namespace
      except AttributeError:
        pass


  #
  # Prepare for running
  #
  def prepare(self, mapper, parameters):

    self._mapper = mapper
    self._parameters = parameters
    self.walk(self.__prepare)

  # Used by prepare()
  def __prepare(self, object):
    if isinstance(object, GRLayoutElement):

      # This all looks confusing, but basically it simply figures out
      # what section a field or summary actually pulls from and also
      # makes sure a section, field, or summary isn't trying to use
      # a source outside of the source's scope.  After it's satisfied
      # with the results, it adds the element to the mapper object.

      # TODO: This can probably be simplified by moving
      # TODO: the various functions to the actual child
      # TODO: object's phaseInits.

      parentSection = object.findParentOfType('GRSection',includeSelf=0)

      if isinstance(object, GRSection) and object._source != None:
        if parentSection is None:
          self._mapper.addSection(object._name, object._source, None)
        else:
          self._mapper.addSection(object._name,
                                  object._source,
                                  parentSection._name)
        object._mymapper = self._mapper.sectionMap[object._name]
        object._mymapper._object = object

      elif isinstance(object, GRField):
        s = parentSection.getAncestorWithSource(object._source)
        if s == None:
          raise GRExceptions.SourceOutOfScope, \
                'Field "%s" in section "%s" uses out-of-scope source "%s"' \
                     % (object.name, parentSection._name, object._source)

        object._section = s._name

        if hasattr(object,'section') and object.section:
          sec = string.lower(object.section)
          curr = s
          section = None
          while curr:
            if curr._name == sec:
              section = curr
              break
            curr = curr._parent.findParentOfType('GRSection')

          if section == None:
            raise GRExceptions.SourceOutOfScope, \
                'Field "%s" in section "%s" uses out-of-scope section "%s"' \
                     % (object.name, parentSection._name, object.section)

          object._section = section._name

        object._mymapper = self._mapper.sectionMap[object._section]

        GDebug.printMesg(6,'Mapping field "%s" to section "%s"' \
                   % (object.name, object._section))
        self._mapper.addFieldToSection(object._section, object.name)

      elif isinstance(object, GRSumm):
        if object.section == None:
          s = object._parent.findParentOfType('GRSection')
        else:
          sec = string.lower(object.section)
          s = self._mapper.sectionMap[sec]._object
        if s == None:
          raise GRExceptions.SourceOutOfScope, \
            'Summary "%s" in section "%s" uses out-of-scope source "%s"' \
                 % (object._field, object._parent.name, object._source)

        object._section = s._name
        object._mymapper = self._mapper.sectionMap[s._name]

        GDebug.printMesg(6,'Mapping summary %s [%s] to section %s' \
                   % (object._field, object.function, object._section))
        self._mapper.addSummaryToSection(object.function,
                                 object._section, object._field)


  #
  # Process layout and dump output.
  # self.prepare *must* be called prior to processing
  #
  def process(self, dest, includeStructuralComments=0):

    global structuralComment
    if includeStructuralComments:
      structuralComment = _structuralComment

    for child in self._children:
      structuralComment(dest, '<!-- [layout] -->')
      child.processAsController(dest, self._mapper)
      structuralComment(dest, '<!-- [/layout] -->\n')




############################################################
#
# Any elements contained within a <layout> element
# inherit GRLayoutElement
#
class GRLayoutElement (GObj,GTriggerExtension):
  def __init__(self, parent=None, type='GRLayoutElement'):
    GTriggerExtension.__init__(self)
    GObj.__init__(self, parent, type=type)
    self.name = None
    self._source = None
    self._mymapper = None
    self._inits = [self.primaryInit]

  # Called if section contains data
  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    return nextSection

  def primaryInit(self):
    if hasattr(self,'source') and self.source is not None:
      self._source = self.source
    else:
      try:
        self._source = self._parent._source
      except AttributeError:
        self._source = None


############################################################
#
# Any elements contained within a <layout> element
# that can themselves be containers
#
class ContainerElement (GRLayoutElement):
  def __init__(self, parent=None, type='GRContainerElement'):
    GRLayoutElement.__init__(self, parent, type=type)
    # TODO: remove.. this is for debugging
    self._name = "untitled"


  # Process all children, handling GRSections properly
  def processChildren(self, dest, mapper,
                      isfirst, islast, firstSection, nextSection):

    first = isfirst or self._type == 'GRSection'

    while 1:
      if isinstance(self, GRSection) :
        self.processTrigger('On-Process')
        
      for child in self._children:
        if child._type == "_content_":
           dest.write(child.getContent())
        else:
          # Handle GRSections specially as
          # they require extra logic
          if isinstance(child, GRSection):
            # If this is the top-most section for a datasource,
            # it needs to be called with processAsController.
            if child._mymapper.toplevel:

              GDebug.printMesg(10, "Calling new controlling section")
              child.processAsController(dest, mapper)

            # ..otherwise call the GRSection's process method.
            else:
                nextSection = child.process(dest, mapper,
                 isfirst=1,
                 islast=(nextSection == None) or \
                    (nextSection._name not in (child._childSections)),
                 firstSection=firstSection,
                 nextSection=nextSection)
          # Don't allow the processing to enter trigger objects
          elif isinstance(child, GTrigger):
            pass
          else:
            # Handle the layout element.
            nextSection = child.process(dest, mapper, first, islast,
                                        firstSection, nextSection)

      first = 0

      if nextSection == self:
        firstSection, nextSection = self.getGotos(mapper)
        islast = (nextSection == None)
      else:
        break

    return nextSection


  #
  #  getAncestorWithSource()
  #
  def getAncestorWithSource(self, source):
    if isinstance(self._parent, GRSection):
      return self._parent.getAncestorWithSource(source)
    else:
      return None



############################################################
#
# <section> tag
#
class GRSection (ContainerElement):

  def __init__(self, parent):
    ContainerElement.__init__(self, parent, 'GRSection')
    self._childSections = []
    self._validTriggers = {'PRE-SECTION':'Pre-Section',
                           'POST-SECTION':'Post-Section',
                           'ON-PROCESS':'On-Process'}

  def _buildObject(self):
    if not hasattr(self,'name') or self.name == None:
      self._name = id(self)
    else:
      self._name = string.lower(self.name)

    # Keep track of "child" section names (for the firstRow stuff)
    psec = self._parent.findParentOfType('GRSection')
    if psec is not None:
      psec._childSections = self._childSections[:]
      psec._childSections.insert(0, self._name)

    return ContainerElement._buildObject(self)


  #
  # Retrieve the first row of data, if available.
  # Handle default data if no rows retrieved.
  #
  def processAsController(self, dest, mapper):

    GDebug.printMesg(10,
       "Controlling section %s's processAsController() called" % self._name)

    if self._source == None:
      # This is a dataless section...
      # simply process it's children

      self.process(dest, mapper, 1, 1, None, None)

    else:
      firstSection, nextSection = self.getGotos(mapper, 1)

      if firstSection:
        # This is a data-bound section and there
        # was data returned, so process section
        self.process(dest, mapper, 1, nextSection == None, self, nextSection)

      else:
        # This is a data-bound section and there
        # was no data returned, so generate default
        # text (if any)

        for c2 in self._children:
          if isinstance (c2, GRDefault):
            c2.processDefault(dest, mapper)



  #
  # Used internally to determine next section to move
  #
  def getGotos(self, mapper, init=0):
    # Determine the next section to process our data.
    # (i.e., handle section grouping)
    if init:
      goto, gotoNext = mapper.getFirstRecord(self._source)
    else:
      goto, gotoNext = mapper.getNextRecord(self._source)

    # Get the actual GRSection objects
    if not goto:
      return (None, None)

    firstSection = mapper.sectionMap[goto]._object
    if gotoNext:
      nextSection = mapper.sectionMap[gotoNext]._object
    else:
      nextSection = None

    # And return them
    return (firstSection, nextSection)


  #
  #  getAncestorWithSource()
  #
  def getAncestorWithSource(self, source):
    if self._source == source:
      return self
    elif isinstance(self._parent, GRSection):
      return self._parent.getAncestorWithSource(source)
    else:
      return None


  # Generic process() method.  Process the
  # current record and handle any children.
  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    self.processTrigger('Pre-Section')
    GDebug.printMesg(10,"Repeating Section %s" % self._name)
    structuralComment(dest,"<!-- [section:%s] -->" % self._name)
    nextSection = self.processChildren(dest, mapper, isfirst, islast, firstSection, nextSection)

    try:

      structuralComment(dest,"<!-- [Setting %s to clear on next set] -->" % self._name)
      mapper.sectionMap[self.name]._clearOnNextSet = 1
    except KeyError:
      pass

    structuralComment(dest,"<!-- [/section:%s] -->" % self._name)
    GDebug.printMesg(10,"Leaving section %s" % self._name)
    self.processTrigger('Post-Section')
    return nextSection


############################################################
#
# <default> tag
# Text in this tag is only displayed if datasource
# returns no data. Note: this will be the only thing
# in the section that will be displayed.
#
class GRDefault (GObj):
  def __init__(self, parent=None, type='GRDefault'):
    GObj.__init__(self, parent, type=type)
    self.name = None
    self._source = None
    self._mymapper = None

  def processDefault(self, dest, mapper):
    structuralComment(dest,"<!-- [default] -->")
    for child in self._children:
      if child._type == "_content_":
        dest.write(child.getContent())
    structuralComment(dest,"<!-- [/default] -->")

  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    return nextSection



############################################################
#
# <field> tag
#
class GRField (GRLayoutElement):
  def __init__(self, parent):
    GRLayoutElement.__init__(self, parent, 'GRField')
    self._section = None
    self.format = None

  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    structuralComment(dest,"<!-- [field:%s] -->" % self.name)
    dest.write (string.replace(saxutils.escape("%s" % self._mymapper.getField(self.name,
                                              self.format)), '\r',''))
    structuralComment(dest,"<!-- [/field:%s] -->" % self.name)
    return nextSection



############################################################
#
# <summ> tag
#
class GRSumm (GRLayoutElement):
  def __init__(self, parent):
    GRLayoutElement.__init__(self, parent, 'GRSumm')
    self.section = None
    self.function = "count"
    self.format = None

  def _buildObject(self):
    if self.section:
      self._section = string.lower(self.section)
    else:
      self._section = None

    self._field = string.lower(self.field)
    return GRLayoutElement._buildObject(self)

  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    structuralComment(dest,"<!-- [summ:%s:%s:%s] -->" % (self.section, self._field, self.function))
    dest.write (self._mymapper.getSummary(self._field, self.function, self.format))
    structuralComment(dest,"<!-- [/summ] -->")
    return nextSection



############################################################
#
# Parameters
#
class GRParam(GRLayoutElement, GRStubParam):
  _typestub = 'GRParam'

  def __init__(self, parent):
    # This is passed parent=None so that
    # the GObj.__init__ doesn't add this
    # object twice to its parent (icky)
    GRLayoutElement.__init__(self, parent=None)
    GRStubParam.__init__(self, parent)

  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    structuralComment(dest,"<!-- [param:%s] -->" % self.name)
    dest.write (self.getFormattedValue())
    structuralComment(dest,"<!-- [/param:%s] -->" % self.name)
    return nextSection




############################################################
#
# Logic support within sections
#

class GRFirstRow(ContainerElement):
  def __init__(self, parent):
    ContainerElement.__init__(self, parent, 'GRFirstRow')

  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    if isfirst:
      nextSection = self.processChildren(dest, mapper, isfirst, islast, firstSection, nextSection)
    return nextSection



class GRNotFirstRow(ContainerElement):
  def __init__(self, parent):
    ContainerElement.__init__(self, parent, 'GRNotFirstRow')

  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    if not isfirst:
      nextSection = self.processChildren(dest, mapper, isfirst, islast, firstSection, nextSection)
    return nextSection


class GRLastRow(ContainerElement):
  def __init__(self, parent):
    ContainerElement.__init__(self, parent, 'GRLastRow')

  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    if islast:
      nextSection = self.processChildren(dest, mapper, isfirst, islast, firstSection, nextSection)
    return nextSection


class GRNotLastRow(ContainerElement):
  def __init__(self, parent):
    ContainerElement.__init__(self, parent, 'GRNotLastRow')

  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    if not islast:
      nextSection = self.processChildren(dest, mapper, isfirst, islast, firstSection, nextSection)
    return nextSection



############################################################
#
# Any namespaced tags designed to
# simply pass thru our layout engine
#
class GRPassThru(ContainerElement):
  def __init__(self, parent):
    ContainerElement.__init__(self, parent, 'GRPassThru')

  def process(self, dest, mapper, isfirst, islast, firstSection, nextSection):
    dest.write('<%s' % self._xmltag)
    for attr in self._loadedxmlattrs.keys():
      dest.write(' %s="%s"' % (attr, saxutils.escape(str(self._loadedxmlattrs[attr]))))

    if len(self._children):
      dest.write('>')
      nextSection = self.processChildren(dest, mapper, isfirst, islast, firstSection, nextSection)
      dest.write('</%s>' % self._xmltag)
    else:
      dest.write('/>')

    return nextSection

  def getXmlTag(self):
    return self._xmltag

